import networkx
try:
    from Creator import Creator
except ImportError:
    from .Creator import Creator


class Solver:
    """
    Class to add edges to given networkx grahps taken from simple Graph Theory,
    such as path, cycle and complete graph.
    """

    @staticmethod
    def path(nxg: networkx.Graph) -> networkx.Graph:
        """
        Adds edges to a given graph as a path, such as the following:
        (0, 1), (1, 2), ... (n-1, n)

        :rtype: networkx.Graph
        :param nxg: A graph with nodes containing coordinates.
        :return: A graph with connected nodes such as they form a path.
        """
        nodes = list(nxg.nodes())

        for edge in zip(nodes[:-1], nodes[1:]):
            x, y = edge
            Creator.add_weighted_edge(nxg, x, y)

        return nxg

    @staticmethod
    def cycle(nxg: networkx.Graph) -> networkx.Graph:
        """
        Adds edges to a given graph as a path, such as the following:
        (0, 1), (1, 2), ... (n-1, n), (n, 0)

        :rtype: networkx.Graph
        :param nxg: A graph with nodes containing coordinates.
        :return: A graph with connected nodes such as they form a cycle.
        """
        # Initially get the path of the graph
        nxg = Solver.path(nxg)

        # Get the start and destination node of the path
        nodes = nxg.nodes()
        start = min(nodes)
        destination = max(nodes)

        # Add the final edge for a cycle
        Creator.add_weighted_edge(nxg, start, destination)

        return nxg

    @staticmethod
    def complete(nxg: networkx.Graph) -> networkx.Graph:
        """
        Makes a graph a complete graph, such as all nodes are connected to
        each other with one edge.

        :rtype: networkx.Graph
        :param nxg: A graph with nodes containing coordinates.
        :return: A complete graph.
        """
        # Get the list of nodes.
        nodes = list(nxg.nodes())
        # Iterate over the nodes as origin
        for origin_id, origin_node in enumerate(nodes):
            # Iterate over the nodes as destination
            for dest_id, dest_node in enumerate(nodes):
                # Discard the nodes that are lower or equal than the origin id
                if dest_id <= origin_id:
                    continue
                # Add an edge between the nodes.
                Creator.add_weighted_edge(nxg, origin_node, dest_node)

        return nxg


# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from os import path
from io import open

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='extended_networkx_tools',
      version='0.7.4.rc1',
      description='Tools package for extending functionality of the networkx package.',
      keywords='graph distributed average consensus convergence rate',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/vonNiklasson/extended-networkx-tools',
      author='Johan Niklasson, Oskar Hahr',
      author_email='jnikl@kth.se, ohahr@kth.se',
      license='MIT',
      packages=find_packages(exclude=['twine']),
      install_requires=[
            'cycler',
            'decorator',
            'kiwisolver',
            'matplotlib',
            'networkx',
            'numpy',
            'pyparsing',
            'python-dateutil',
      ],
      py_modules=['six'],
      python_requires='~=3.0',
      zip_safe=False,
      classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',

            # Indicate who your project is intended for
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',

            # Topics
            'Topic :: System :: Distributed Computing',
            'Topic :: Scientific/Engineering :: Mathematics',


            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: MIT License',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.2',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
      ],
)



import queue
from typing import List, Dict

import networkx as nx
import numpy as np
from numpy import linalg


class Analytics:

    @staticmethod
    def get_neighbour_matrix(nxg: nx.Graph) -> List[List[float]]:
        """
        Creates a neighbour matrix for a specified graph: g, each row represents a node in the graph
        where the values in each column represents if there is an edge or not between those nodes.

        :param nxg: networkx bi-directional graph object.
        :type nxg: nx.Graph
        :return A: List of rows, representing the adjacency matrix.
        :rtype: List[List[float]]
        """
        # Sort the nodes in the graph
        s_nodes = list(nxg.nodes())
        s_nodes.sort()
        # Get the dimension of each row
        dim = len(s_nodes)

        mx = []
        for node in nxg.nodes():
            row = [0] * dim
            # Get the index of the current node
            node_index = s_nodes.index(node)
            row[node_index] = 1
            for neighbour in nxg.neighbors(node):
                node_index = s_nodes.index(neighbour)
                row[node_index] = 1
            mx.append(row)
        return mx

    @staticmethod
    def get_stochastic_neighbour_matrix(nxg: nx.Graph) -> List[List[float]]:
        """
        Creates a stochastic adjacency matrix for a specified graph: g, each row represents a node in the graph
        where the values in each column represents if there is an edge or not between those nodes.
        The values for each neighbour is represented by 1/(number of neighbours), if no edge exists this value is 0.

        :param nxg: Networkx bi-directional graph object.
        :type nxg: nx.Graph
        :return A: List of rows, representing the adjacency matrix.
        :rtype: List[List[float]]
        """
        # Get the neighbour matrix
        mx = Analytics.get_neighbour_matrix(nxg)

        # Iterate over each row
        for row_id, _ in enumerate(mx):
            # Calculate the sum for each row
            row_sum = sum(mx[row_id])
            # Divide each node in the row with the sum of the row
            mx[row_id] = list(map(lambda x: (x / row_sum), mx[row_id]))

        # Working solution that might however be worse than the previous solution.
        # mx = list(map(lambda row: list(map(lambda cell: cell / sum(row), row)), mx))

        return mx

    @staticmethod
    def get_eigenvalues(a: List[List[float]]) -> List[float]:
        """
        Simple function to retrieve the eigenvalues of a matrix.

        :param a: A matrix made up of nested lists.
        :return: List of eigenvalues of the provided matrix.
        :rtype: List[float]
        """
        
        A = np.array(a)
        w, _ = linalg.eig(A)
        return w

    @staticmethod
    def second_largest(numbers: List[float]) -> float:
        """
        Simple function to return the 2nd largest number in a list of numbers.

        :param numbers: A list of numbers
        :return: The 2nd largest number in the list numbers
        :rtype: float

        """
        count = 0
        m1 = m2 = float('-inf')
        for x in numbers:
            count += 1
            if x > m2:
                if x >= m1:
                    m1, m2 = x, m1
                else:
                    m2 = x
        return m2 if count >= 2 else None

    @staticmethod
    def convergence_rate(nxg: nx.Graph) -> float:
        """
        Function to retrieve the 2nd largest eigenvalue in the adjacency matrix of a graph

        :param nxg: networkx bi-directional graph object
        :type nxg: nx.Graph
        :return: The 2nd largest eigenvalue of the adjacency matrix
        :rtype: float
        """
        A = Analytics.get_stochastic_neighbour_matrix(nxg)
        ev = Analytics.get_eigenvalues(A)
        return Analytics.second_largest(ev)

    @staticmethod
    def total_edge_cost(nxg: nx.Graph) -> float:
        """
        Calculates the total cost of all edges in the given graph

        :param nxg: A networkx object with nodes and edges.
        :type nxg: nx.Graph
        :return: The total cost of all edges in the graph.
        :rtype: float
        """
        total = 0
        edges = nxg.edges(data=True)

        for edge in edges:
            if 'weight' in edge[2]:
                total += edge[2]['weight']
            
        return total

    @staticmethod
    def get_distance_distribution(nxg: nx.Graph) -> Dict[int, int]:
        """
        Makes a list representing the distribution of longest shortest paths between every node
        in the graph.

        :rtype: Dict[int, int]
        :param nxg: A given graph with edges.
        :return: A dict with a distribution of the longest shortest paths between nodes.
        """
        # Get a list of all paths
        paths = list(nx.networkx.all_pairs_shortest_path_length(nxg))
        # Create an empty dict of distance distributions
        distributions = {}
        # Iterate over each path
        for origin, path in paths:
            # Make sure we don't check the same path twice
            max_node_distance = -1
            for dest in range(0, len(path)):
                # Get the actual shortest distance between 2 nodes
                max_node_distance = max(max_node_distance, path.get(dest))
                # Make sure we create the distance first, then add one to it
            if max_node_distance is not -1:
                if max_node_distance not in distributions:
                    distributions[max_node_distance] = 1
                else:
                    distributions[max_node_distance] += 1

        return distributions

    @staticmethod
    def is_nodes_connected(nxg: nx.Graph, origin: int, destination: int) -> bool:
        """
        Checks if two nodes are connected with each other using a BFS approach.

        :param nxg: The grapg that contains the two nodes.
        :param origin: The origin node id to check from.
        :param destination: The destination node to check the connectivity to.
        :return: True if there's a connection between the nodes, otherwise False.
        """
        # Create a set of seen nodes
        seen = set()
        # Create a queue
        q = queue.Queue()

        # Add the start node
        q.put(origin)

        # Iterate while the queue is not empty
        while not q.empty():
            # Get the first element of the queue
            node_id = q.get()
            # If we're at the destination, return True
            if node_id == destination:
                return True

            # Otherwise, add the node to seen
            seen.add(node_id)
            # Iterate over the neighbours, but discard the first element
            # since it should be the origin
            for _, neighbour in nxg.edges(node_id):
                # If we haven't seen it, add it to the queue
                if neighbour not in seen:
                    q.put(neighbour)

        # If we've reached here we haven't found the other node. If so, return False
        return False



from typing import Dict, Set, List, Tuple

import networkx
from random import randint


class Creator:
    """
    Static class that works with creating graph objects from given specifications.
    Can either create a random unassigned graph with given nodes or a graph with
    edges from given parameters.
    """

    @staticmethod
    def from_random(node_count: int) -> networkx.Graph:
        """
        Creates an unassigned graph with nodes of random position.
        The work area corresponds to the node count squared.

        :rtype: networkx.Graph
        :param node_count: The number of nodes to create a graph from.
        :return: An unassigned graph with nodes with random position.
        """
        area_dimension = node_count
        nxg: networkx.Graph = networkx.Graph()

        node_set: Set[str] = set()

        for node_id in range(0, node_count):
            coord_x = None
            coord_y = None
            # Keep finding new coordinates until a new one is found
            while coord_x is None or coord_y is None or str((coord_x, coord_y)) in node_set:
                coord_x = randint(0, area_dimension)
                coord_y = randint(0, area_dimension)
            nxg.add_node(node_id, x=coord_x, y=coord_y)
            node_set.add(str((coord_x, coord_y)))

        return nxg

    @staticmethod
    def from_spec(v: Dict[int, Tuple[int, int]], e: Dict[int, List[int]]) -> networkx.Graph:
        """
        Creates a graph from given parameters, that also assigns weighted edges based on a neighbour list.

        :param v: Nodes in the graph. Should be a dict with the format
                {  node_1: (x, y), node_2: (x, y)... }
        :param e: Edges that connects the nodes.Should be a dict with the format
                { node_1: [dest_1, dest_2, ...], node_2: [dest_3, dest_4, ...] }
        :return: A graph with assigned nodes and weighted edges.
        :rtype: networkx.Graph
        """
        nxg = networkx.Graph()

        for node_id, node in v.items():
            nxg.add_node(node_id, x=node[0], y=node[1])

        for origin, destinations in e.items():
            for destination in destinations:
                Creator.add_weighted_edge(nxg, origin, destination)

        return nxg

    @staticmethod
    def add_weighted_edge(nxg: networkx.Graph, origin: int, destination: int) -> bool:
        """
        Adds a bidirectional edge between 2 nodes with weight corresponding to the
        distance between the nodes squared.

        :param nxg: The graph to add an edge to.
        :param origin: First node id to add the edge from
        :param destination: Second node id to add the edge to.
        :return: True if the edge was added, otherwise false if the edge already existed.
        """
        if nxg.has_edge(origin, destination):
            return False

        # Find start and end coordinates
        start_coord = (nxg.node[origin]['x'], nxg.node[origin]['y'])
        destination_coord = (nxg.node[destination]['x'], nxg.node[destination]['y'])

        # Subtract the coordinate values of the two points
        delta = tuple(map(lambda n: pow(n[0] - n[1], 2), zip(start_coord, destination_coord)))
        # Extract values from delta tuple
        delta_x, delta_y = delta
        # Cost is the summation of the difference in x and y of the two coordinates
        weight = delta_x + delta_y

        # Add edge to graph with its corresponding weight
        nxg.add_edge(origin, destination, weight=weight)

        return True
